// $Id: BroString.cc,v 1.3 2005/10/05 03:38:57 vern Exp $
//
// Copyright (c) 1997, 1998, 1999, 2002, 2003
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#include "config.h"

#include <algorithm>
#include <ctype.h>

#include "BroString.h"
#include "Var.h"

#ifdef DEBUG
#define DEBUG_STR(msg) DBG_LOG(DBG_STRING, msg)
#else
#define DEBUG_STR(msg)
#endif


// This constructor forces the user to specify arg_final_NUL.  When str
// is a *normal* NUL-terminated string, make arg_n == strlen(str) and
// arg_final_NUL == 1; when str is a sequence of n bytes, make
// arg_final_NUL == 0.

BroString::BroString(int arg_final_NUL, byte_vec str, int arg_n)
	{
	b = str;
	n = arg_n;
	final_NUL = arg_final_NUL;
	use_free_to_delete = 0;
	}

BroString::BroString(const u_char* str, int arg_n, int add_NUL)
	{
	n = arg_n;
	b = new u_char[add_NUL ? n + 1 : n];
	memcpy(b, str, n);

	if ( add_NUL )
		{
		b[n] = 0;
		final_NUL = 1;
		}
	else
		final_NUL = 0;

	use_free_to_delete = 0;
	}

BroString::BroString(const char* str)
	{
	n = strlen(str);
	b = new u_char[n+1];
	memcpy(b, str, n+1);
	final_NUL = 1;
	use_free_to_delete = 0;
	}

BroString::BroString(const string &str)
	{
	n = str.size();
	b = new u_char[n+1];
	memcpy(b, str.c_str(), n+1);
	final_NUL = 1;
	use_free_to_delete = 0;
	}

BroString::BroString(const BroString& bs)
	{
	b = 0;
	n = 0;
	final_NUL = 0;
	*this = bs;
	}

void BroString::Reset()
	{
	if ( use_free_to_delete )
		free(b);
	else
		delete [] b;

	b = 0;
	n = 0;
	final_NUL = 0;
	}

const BroString& BroString::operator=(const BroString &bs)
	{
	Reset();
	n = bs.n;
	b = new u_char[n+1];

	memcpy(b, bs.b, n);
	b[n] = '\0';

	final_NUL = 1;
	use_free_to_delete = 0;
	return *this;
	}

bool BroString::operator==(const BroString &bs) const
	{
	bool result = Bstr_eq(this, &bs);
	DEBUG_STR(fmt("%s == %s: %i", ExpandedString(),
			bs.ExpandedString(), result));
	return result;
	}

bool BroString::operator<(const BroString &bs) const
	{
	bool result = Bstr_cmp(this, &bs) < 0;
	DEBUG_STR(fmt("%s < %s: %i", ExpandedString(),
			bs.ExpandedString(), result));
	return result;
	}

const char* BroString::CheckString() const
	{
	if ( n == 0 )
		return "";

	if ( memchr(b, '\0', n + final_NUL) != &b[n] )
		{
		// Either an embedded NUL, or no final NUL.
		char* exp_s = ExpandedString();
		if ( b[n-1] != '\0' )
			run_time("string without NUL terminator: \"%s\"", exp_s);
		else
			run_time("string with embedded NUL: \"%s\"", exp_s);

		delete [] exp_s;
		return "<string-with-NUL>";
		}

	return (const char*) b;
	}

char* BroString::ExpandedString(int format) const
	{
	// Maxmimum character expansion is as \xHH, so a factor of 4.
	char* s = new char[n*4 + 1];	// +1 is for final '\0'
	char* sp = s;
	for ( int i = 0; i < n; ++i )
		{
		if ( b[i] == '\0' && format == EXPANDED_STRING )
			{
			// Note, generating "\0" for BRO_STRING_LITERAL
			// would cause ambiguity in cases like "\021".
			*sp++ = '\\'; *sp++ = '0';
			}

		else if ( b[i] == '\x7f' && format == EXPANDED_STRING )
			{
			*sp++ = '^'; *sp++ = '?';
			}

		else if ( b[i] <= 26 && format == EXPANDED_STRING )
			{
			*sp++ = '^'; *sp++ = b[i] + 'A' - 1;
			}

		else if ( (b[i] == '\\' || b[i] == '"') &&
			  format == BRO_STRING_LITERAL )
			{
			*sp++ = '\\'; *sp++ = b[i];
			}

		else if ( b[i] >= ' ' && b[i] < 127 )
			*sp++ = b[i];

		else
			{
			char hex_fmt[16];

			*sp++ = '\\'; *sp++ = 'x';
			sprintf(hex_fmt, "%02x", b[i]);
			*sp++ = hex_fmt[0]; *sp++ = hex_fmt[1];
			}
		}

	*sp = '\0';	// NUL-terminate.

	return s;
	}

void BroString::ToUpper()
	{
	for ( int i = 0; i < n; ++i )
		if ( islower(b[i]) )
			b[i] = toupper(b[i]);
	}

BroString* BroString::GetSubstring(int start, int len) const
	{
	// This code used to live in bro.bif's sub_bytes() routine.
	if ( start < 0 || start > n )
		return 0;

	if ( len < 0 || len > n - start )
		len = n - start;

	return new BroString(&b[start], len, 1);
	}

int BroString::FindSubstring(const BroString* s) const
	{
	return strstr_n(n, b, s->Len(), s->Bytes());
	}

BroString::Vec* BroString::Split(const BroString::IdxVec& indices) const
	{
	unsigned int i;

	if ( indices.size() == 0 )
		return 0;

	// Copy input, ensuring space for "0":
	IdxVec idx(1 + indices.size());

	idx[0] = 0;
	idx.insert(idx.end(), indices.begin(), indices.end());

	// Sanity checks.
	for ( i = 0; i < idx.size(); ++i )
		if ( idx[i] >= n || idx[i] < 0 )
			idx[i] = 0;

	// Sort it:
	sort(idx.begin(), idx.end());

	// Shuffle vector so duplicate entries are used only once:
	IdxVecIt end = unique(idx.begin(), idx.end());

	// Each element in idx is now the start index of a new
	// substring, and we know that all indices are within [0, n].
	//
	Vec* result = new Vec();
	int last_idx = -1;
	int next_idx;
	i = 0;

	for ( IdxVecIt it = idx.begin(); it != end; ++it, ++i )
		{
		int len = (it + 1 == end) ? -1 : idx[i + 1] - idx[i];
		result->push_back(GetSubstring(idx[i], len));
		}

	return result;
	}

VectorVal* BroString:: VecToPolicy(Vec* vec)
	{
	VectorVal* result =
		new VectorVal(internal_type("string_vec")->AsVectorType());
	if ( ! result )
		return 0;

	for ( unsigned int i = 0; i < vec->size(); ++i )
		{
		BroString* string = (*vec)[i];
		StringVal* val = new StringVal(string->Len(),
						(const char*) string->Bytes());
		result->Assign(i+1, val, 0);
		}

	return result;
	}

BroString::Vec* BroString::VecFromPolicy(VectorVal* vec)
	{
	Vec* result = new Vec();

	// VectorVals start at index 1!
	for ( unsigned int i = 1; i <= vec->Size(); ++i )
		{
		Val* v = vec->Lookup(i);	// get the RecordVal
		if ( ! v )
			continue;

		BroString* string = new BroString(*(v->AsString()));
		result->push_back(string);
		}

	return result;
	}

char* BroString::VecToString(const Vec* vec)
	{
	string result("[");

	for ( BroString::VecCIt it = vec->begin(); it != vec->end(); ++it )
		{
		result += (*it)->ExpandedString();
		result += ",";
		}

	result += "]";

	return strdup(result.c_str());
	}

int Bstr_eq(const BroString* s1, const BroString* s2)
	{
	if ( s1->Len() != s2->Len() )
		return 0;

	return memcmp(s1->Bytes(), s2->Bytes(), s1->Len()) == 0;
	}

int Bstr_cmp(const BroString* s1, const BroString* s2)
	{
	int n = min(s1->Len(), s2->Len());
	int cmp = memcmp(s1->Bytes(), s2->Bytes(), n);

	if ( cmp || s1->Len() == s2->Len() )
		return cmp;

	// Compared equal, but one was shorter than the other.  Treat
	// it as less than the other.
	if ( s1->Len() < s2->Len() )
		return -1;
	else
		return 1;
	}

BroString* concatenate(std::vector<data_chunk_t>& v)
	{
	int n = v.size();
	int len = 0;
	int i;
	for ( i = 0; i < n; ++i )
		len += v[i].length;

	char* data = new char[len+1];

	char* b = data;
	for ( i = 0; i < n; ++i )
		{
		memcpy(b, v[i].data, v[i].length);
		b += v[i].length;
		}

	*b = '\0';

	return new BroString(1, (byte_vec) data, len);
	}

BroString* concatenate(BroString::CVec& v)
	{
	int n = v.size();
	int len = 0;
	int i;
	for ( i = 0; i < n; ++i )
		len += v[i]->Len();

	char* data = new char[len+1];

	char* b = data;
	for ( i = 0; i < n; ++i )
		{
		memcpy(b, v[i]->Bytes(), v[i]->Len());
		b += v[i]->Len();
		}
	*b = '\0';

	return new BroString(1, (byte_vec) data, len);
	}

BroString* concatenate(BroString::Vec& v)
	{
	BroString::CVec cv;

	for ( BroString::VecIt it = v.begin(); it != v.end(); ++it )
		cv.push_back(*it);

	return concatenate(cv);
	}

void delete_strings(std::vector<const BroString*>& v)
	{
	for ( unsigned int i = 0; i < v.size(); ++i )
		delete v[i];
	v.clear();
	}
